from functools import partial
from unittest import TestCase
from uuid import uuid4

from pydantic.types import conint

from openmodule.models.base import OpenModuleModel
from openmodule.rpc import server
from openmodule.rpc.common import channel_to_response_topic, channel_to_request_topic
from openmodule.rpc.server import RPCServer, RPCRequest
from openmodule_test.rpc import RPCServerTestMixin
from openmodule_test.zeromq import ZMQTestMixin


def _fake_request(gate=..., direction=...):
    gateway = {}
    if gate != Ellipsis:
        gateway["gate"] = gate
    if direction != Ellipsis:
        gateway["direction"] = direction

    request = {}
    if gateway:
        request["gateway"] = gateway

    return request


class RPCCommonTestCase(TestCase):
    def test_topics(self):
        self.assertEqual(b"rpc-rep-test", channel_to_response_topic(b"test"))
        self.assertEqual(b"rpc-req-test", channel_to_request_topic(b"test"))

    def test_gateway_filter(self):
        filter_full = server.gateway_filter("gate1", "in")
        self.assertTrue(filter_full(_fake_request(gate="gate1", direction="in"), None, None))
        self.assertFalse(filter_full(_fake_request(gate="gate1", direction="out"), None, None))
        self.assertFalse(filter_full(_fake_request(gate="gate2", direction="in"), None, None))
        self.assertFalse(filter_full(_fake_request(gate="gate2", direction="out"), None, None))
        self.assertFalse(filter_full(_fake_request(gate="gate2"), None, None))
        self.assertFalse(filter_full(_fake_request(gate="gate2"), None, None))
        self.assertFalse(filter_full(_fake_request(direction="in"), None, None))
        self.assertFalse(filter_full(_fake_request(direction="out"), None, None))
        self.assertFalse(filter_full(_fake_request(), None, None))

        filter_direction = server.gateway_filter(direction="in")
        self.assertTrue(filter_direction(_fake_request(gate="gate1", direction="in"), None, None))
        self.assertFalse(filter_direction(_fake_request(gate="gate1", direction="out"), None, None))
        self.assertTrue(filter_direction(_fake_request(gate="gate2", direction="in"), None, None))
        self.assertFalse(filter_direction(_fake_request(gate="gate2", direction="out"), None, None))
        self.assertFalse(filter_direction(_fake_request(gate="gate2"), None, None))
        self.assertFalse(filter_direction(_fake_request(gate="gate2"), None, None))
        self.assertTrue(filter_direction(_fake_request(direction="in"), None, None))
        self.assertFalse(filter_direction(_fake_request(direction="out"), None, None))
        self.assertFalse(filter_direction(_fake_request(), None, None))

        filter_gate = server.gateway_filter(gate="gate1")
        self.assertTrue(filter_gate(_fake_request(gate="gate1", direction="in"), None, None))
        self.assertTrue(filter_gate(_fake_request(gate="gate1", direction="out"), None, None))
        self.assertFalse(filter_gate(_fake_request(gate="gate2", direction="in"), None, None))
        self.assertFalse(filter_gate(_fake_request(gate="gate2", direction="out"), None, None))
        self.assertFalse(filter_gate(_fake_request(gate="gate2"), None, None))
        self.assertFalse(filter_gate(_fake_request(gate="gate2"), None, None))
        self.assertFalse(filter_gate(_fake_request(direction="in"), None, None))
        self.assertFalse(filter_gate(_fake_request(direction="out"), None, None))
        self.assertFalse(filter_gate(_fake_request(), None, None))


class TestRPCRequest(OpenModuleModel):
    __test__ = False
    pass


class TestRPCResponse(OpenModuleModel):
    __test__ = False
    pass


class TestRPCResponse2(OpenModuleModel):
    __test__ = False
    some_payload: str


class RPCServerTestCase(RPCServerTestMixin, ZMQTestMixin, TestCase):
    rpc_channels = ["channel", "channel2"]

    server: RPCServer

    def setUp(self):
        super().setUp()
        self.called_types = {}
        self.server = RPCServer(self.zmq_context(), self.zmq_config())
        self.server_thread = self.server.run_as_thread()

    def tearDown(self):
        self.server.shutdown()
        self.server_thread.join()
        super().tearDown()

    def set_called(self, request, message: RPCRequest, value=True):
        self.called_types[message.type] = value

    def set_called_value(self, value):
        return partial(self.set_called, value=value)

    def test_invalid_rpc_request(self):
        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)

        rpc_request = RPCRequest(name="testclient", type="test-type", rpc_id=uuid4(), request={})
        broken_data = rpc_request.dict()
        del broken_data["rpc_id"]
        self.zmq_client.send(f"rpc-req-channel".encode("ascii"), broken_data)

        _, error_response = self.zmq_client.wait_for_message(
            filter=lambda topic, message: topic == b"rpc-rep-channel" and message.get("type") == "unknown"
        )
        self.assertIsNone(error_response.get("rpc_id"))

    def test_no_filter(self):
        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)

        self.rpc("channel", "test-type", {})
        self.assertTrue(self.called_types.get("test-type"))

    def test_exception_in_filter_function(self):
        def bad_filter(*args, **kwargs) -> bool:
            raise Exception("Error123")

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)
        self.server.add_filter(bad_filter)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCFailure(response, "filter_error")
        self.assertEqual("Error123", response["response"]["exception"])

    def test_filter(self):
        def allow_only_type_test(message: RPCRequest, **kwargs) -> bool:
            return message.type == "test-type"

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCSuccess(response)
        self.assertTrue(self.called_types.get("test-type"))

        response = self.rpc("channel", "test-type2", {})
        self.assertTrue(self.called_types.get("test-type2"))
        self.assertRPCSuccess(response)

        # clear stats, and add the filter
        del self.called_types["test-type"]
        del self.called_types["test-type2"]
        self.server.add_filter(allow_only_type_test)

        # should still work as before
        response = self.rpc("channel", "test-type", {})
        self.assertRPCSuccess(response)
        self.assertTrue(self.called_types.get("test-type"))

        # times out
        with self.assertRaises(TimeoutError):
            self.rpc("channel", "test-type2", {})
        self.assertNotIn("test-type2", self.called_types)

    def test_filter_channel_type(self):
        def f(*args, **kwargs):
            return False

        def check_ok():
            response = self.rpc("channel", "test-type", {})
            self.assertRPCSuccess(response)
            self.assertTrue(self.called_types.get("test-type"))
            self.called_types.clear()

        def check_not_called():
            with self.assertRaises(TimeoutError):
                response = self.rpc("channel", "test-type", {})
            self.assertEqual(None, self.called_types.get("test-type"))
            self.called_types.clear()

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)

        self.server.add_filter(f, "abc")
        check_ok()

        self.server.add_filter(f, "channel", "abc")
        check_ok()

        self.server.add_filter(f, "channel")
        check_not_called()

        self.server.filters.clear()
        self.server.add_filter(f, "channel", "test-type")
        check_not_called()

    def test_can_only_register_once(self):
        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.wait_for_rpc_server(self.server)
        with self.assertRaises(ValueError) as e:
            self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
        self.assertIn("already registered", str(e.exception))

    def test_exception_in_handler(self):
        def handler(*_, **__):
            raise Exception("Error123")

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCFailure(response, "handler_error")
        self.assertEqual("Error123", response["response"]["exception"])

    def test_handler_returns_none(self):
        def handler(*_, **__):
            return None

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCSuccess(response)

    def test_handler_returns_dict(self):
        def handler(*_, **__):
            return {}

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCSuccess(response)

    def test_handler_returns_model(self):
        def handler(*_, **__):
            return TestRPCResponse()

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})
        self.assertRPCSuccess(response)

    def test_handler_returns_wrong_dict(self):
        def bad_handler(*_, **__):
            return {}

        def good_handler(*_, **__):
            return {"some_payload": "test"}

        self.server.register_handler("channel", "test-type1", TestRPCRequest, TestRPCResponse2, bad_handler)
        self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse2, good_handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type1", {})
        self.assertRPCFailure(response, "handler_error")

        response = self.rpc("channel", "test-type2", {})
        self.assertRPCSuccess(response)

    def test_handler_returns_wrong_response_type(self):
        def bad_handler(*_, **__):
            return TestRPCResponse()

        def good_handler(*_, **__):
            return TestRPCResponse2(some_payload="test")

        self.server.register_handler("channel", "test-type1", TestRPCRequest, TestRPCResponse2, bad_handler)
        self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse2, good_handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type1", {})
        self.assertRPCFailure(response, "handler_error")

        response = self.rpc("channel", "test-type2", {})
        self.assertRPCSuccess(response)

    def test_handler_returns_additional_data(self):
        """
        additional data is dropped, just as the serializer dictates
        """

        def too_much_handler(*_, **__):
            return {"some_payload": "test", "nobody_wants_you": ":("}

        self.server.register_handler("channel", "test-type1", TestRPCRequest, TestRPCResponse2, too_much_handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type1", {})
        self.assertRPCSuccess(response)
        self.assertDictEqual(response["response"], {"some_payload": "test", "status": "ok"})

    def test_serializer_error(self):
        class SomeValidator(OpenModuleModel):
            max_int: conint(gt=0, lt=10)

        def handler(*_, **__):
            pass

        self.server.register_handler("channel", "test-type", SomeValidator, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        self.assertRPCSuccess(self.rpc("channel", "test-type", {"max_int": 1}))
        self.assertRPCSuccess(self.rpc("channel", "test-type", {"max_int": 5}))
        self.assertRPCFailure(self.rpc("channel", "test-type", {"max_int": 0}), "validation_error")
        self.assertRPCFailure(self.rpc("channel", "test-type", {"max_int": 10}), "validation_error")

    def test_no_handler_found(self):
        def handler(*_, **__):
            return TestRPCResponse()

        self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
        self.wait_for_rpc_server(self.server)

        response = self.rpc("channel", "test-type", {})  # correct type and channel
        self.assertRPCSuccess(response)

        with self.assertRaises(TimeoutError):  # wrong type
            self.rpc("channel", "test-type-doesnotexist", {})

        with self.assertRaises(TimeoutError):  # wrong channel (correct prefix chosen on purpose)
            self.rpc("channel2", "test-type", {})

        with self.assertRaises(TimeoutError):  # wrong everything
            self.rpc("channel2", "test-type-doesnotexist", {})



from openmodule_test.zeromq import ZMQTestMixin


class AlertTestMixin(ZMQTestMixin):
    def assertAlert(self, timeout=3, **alert_kwargs):
        self.assertSubscription(b"alert")
        while True:
            message = self.zmq_client.wait_for_message_on_topic(topic=b"alert", timeout=timeout)
            if all(message.get(k) == v for k, v in alert_kwargs.items()):
                return message

    def assertNoAlert(self, timeout=3, **alert_kwargs):
        with self.assertRaises(TimeoutError):
            self.assertAlert(timeout, **alert_kwargs)



import logging
import time
from collections import namedtuple
from enum import Enum
from typing import Optional, Dict, Callable, Any

from openmodule.models.base import ZMQMessage, OpenModuleModel

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from openmodule.core import OpenModuleCore


class HealthzStatus(str, Enum):
    ok = "ok"
    error = "error"


class HealthPongPayload(OpenModuleModel):
    status: HealthzStatus
    meta: Dict
    version: str
    name: str
    startup: float
    message: Optional[str]


class HealthPongMessage(ZMQMessage):
    type: str = "healthz"
    pong: HealthPongPayload


class HealthPingMessage(ZMQMessage):
    type: str = "ping"


startup = time.time()

HealthResult = namedtuple("HealthResult", "status,message,meta")


def health_ok(meta=None) -> HealthResult:
    return HealthResult(HealthzStatus.ok, None, meta)


def health_error(message, meta=None) -> HealthResult:
    return HealthResult(HealthzStatus.error, message, meta)


HealthHandlerType = Callable[[], HealthResult]


class Healthz:
    health_handler: HealthHandlerType

    def __init__(self, core: 'OpenModuleCore', health_handler: Optional[HealthHandlerType] = None):
        self.health_handler = health_handler or health_ok
        self.core = core
        self.log = logging.getLogger(self.__class__.__name__)

    def process_message(self, _: HealthPingMessage):
        try:
            result = self.health_handler()
        except Exception as e:
            self.log.exception(e)
            result = health_error("internal error in health routine")

        if not isinstance(result, HealthResult):
            self.log.error(f"invalid return type '{type(result)}' from health handler, "
                           f"expected HealthResult")
            result = health_error("internal error in health routine")

        response = HealthPongMessage(
            name=self.core.config.NAME,
            pong=dict(
                status=result.status,
                message=result.message,
                version=self.core.config.VERSION,
                name=self.core.config.NAME,
                startup=startup,
                meta=result.meta or {}))
        self.core.publish(response, b"healthz")



from sqlalchemy.orm import Session


def update_query(session: Session, query, values: dict):
    """
    in order to update via a query in sqlite we have to no synchronize the session
    after this operation all db objects are expired (same as if you access the object
    after the transaction)
    """
    res = query.update(values, synchronize_session=False)
    session.expire_all()
    return res


def delete_query(session: Session, query):
    """
    in order to delete via a query in sqlite we have to not synchronize the session
    after this operation all db objects are expired (same as if you access the object
    after the transaction)
    """
    res = query.delete(synchronize_session=False)
    session.expire_all()
    return res



import os
import threading

from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine, event, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def alembic_config(connection, alembic_path):
    alembic_cfg = Config(os.path.join(alembic_path, "alembic.ini"),
                         attributes={
                             "configure_logging": False,
                             "connection": connection,
                         })
    alembic_cfg.set_main_option("script_location", os.path.join(alembic_path, "alembic"))
    return alembic_cfg


def migrate_database(connection, alembic_path=None):
    if alembic_path is None:
        alembic_path = os.path.join(os.getcwd(), "database")
    assert os.path.exists(os.path.abspath(alembic_path)), f"alembic path {os.path.abspath(alembic_path)} does not exist"
    config = alembic_config(connection, alembic_path)
    command.upgrade(config, "head")


active_databases = {}

target_metadata = None


def register_bases(bases):
    global target_metadata
    target_metadata = MetaData()

    if not isinstance(bases, list):
        bases = [bases]

    for base in bases:
        for table in base.metadata.tables.values():
            table.tometadata(target_metadata)


def database_name(db_folder, db_name):
    return os.path.join(db_folder, db_name) + ".sqlite3"


def get_database(db_folder: str, name: str, alembic_path=None):
    global active_databases
    tmp = database_name(db_folder, name)
    assert active_databases.get(tmp) is None, f"database {tmp} already exists," \
                                              f" check if it was shutdown before a new one was created"
    os.makedirs(db_folder, exist_ok=True)
    path = f"sqlite:///{tmp}"
    engine = create_engine(path, poolclass=StaticPool, connect_args={'check_same_thread': False})
    migrate_database(engine, alembic_path)
    active_databases[tmp] = engine
    return engine


class Database:
    def __init__(self, database_folder, name="database", alembic_path=None, expire_on_commit=True):
        self.db_folder = database_folder
        self.name = name
        self._engine = get_database(database_folder, name, alembic_path)
        self._session = sessionmaker(self._engine)
        self.active_session = None
        self.expire_on_commit = expire_on_commit
        self.lock = threading.RLock()

    def __del__(self):
        active_databases.pop(database_name(self.db_folder, self.name), None)

    def shutdown(self):
        assert self._session, (
            "database is already closed, you called shutdown twice somewhere"
        )
        with self.lock:
            self._session = None
            global active_databases
            active_databases.pop(database_name(self.db_folder, self.name), None)

    def __enter__(self):
        assert self._session, "Database is already closed"
        self.lock.acquire()
        self.active_session = self._session(expire_on_commit=self.expire_on_commit)
        return self.active_session

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.active_session.commit()
            else:
                self.active_session.rollback()
            self.active_session.close()
        except Exception as e:
            exc_type = exc_type or type(e)
        finally:
            self.active_session = None
            self.lock.release()
        return exc_type is None



import random
import string
import time
from contextlib import suppress
from unittest import TestCase

import orjson
import zmq

from openmodule.messaging import get_pub_socket, get_sub_socket, receive_message_from_socket
from openmodule_test.zeromq import TestBroker, fake_config, TestClient


class MessagingTest(TestCase):
    pub_socket: zmq.Socket
    sub_socket: zmq.Socket
    broker: TestBroker

    def setUp(self) -> None:
        super().setUp()

        # we cannot use ZMQTestMixin here, because the zmq client crashes (on purpose) when receiving invalid messages

        random_suffix = "".join(random.choices(string.ascii_letters, k=10))
        sub = f"inproc://test-sub-{random_suffix}"
        pub = f"inproc://test-pub-{random_suffix}"
        self.broker = TestBroker(sub, pub)
        self.broker.start()

        self.pub_socket = get_pub_socket(self.broker.context, fake_config(self.broker))
        self.sub_socket = get_sub_socket(self.broker.context, fake_config(self.broker))
        self.sub_socket.subscribe(b"topic")

        connected = False
        for x in range(30):
            self.pub_socket.send_multipart((b"topic", b"data"))
            time.sleep(0.1)
            try:
                self.sub_socket.recv_multipart(flags=zmq.NOBLOCK)
                connected = True
                break
            except zmq.Again:
                pass

        # sometimes a stray message is received afterwards, we have to deal with them here in quite a "dumb" way
        # the test client handles this much better because it can differentiate between connect pings and actual
        # messages
        with suppress(zmq.Again): # pragma: no cover
            self.sub_socket.recv_multipart(flags=zmq.NOBLOCK)
            time.sleep(0.1)
            self.sub_socket.recv_multipart(flags=zmq.NOBLOCK)

        assert connected, "sockets did not connect"

    def tearDown(self) -> None:
        self.pub_socket.close()
        self.sub_socket.close()
        self.broker.stop()
        super().tearDown()

    def test_test_client_raises_on_invalid_message(self):
        client = TestClient(self.broker)
        client.start()
        client.subscribe(b"topic")

        self.pub_socket.send_multipart((b"topic", b"many", b"parts"))
        with self.assertRaises(TimeoutError):
            client.wait_for_message(filter=lambda x: True)

        self.assertFalse(client.is_alive())

    def test_single_part(self):
        self.pub_socket.send_multipart((b"topic",))
        with self.assertLogs() as cm:
            res = receive_message_from_socket(self.sub_socket)
        self.assertEqual(res, (None, None))
        self.assertIn("invalid number of parts: 1", str(cm.output))

    def test_three_parts(self):
        self.pub_socket.send_multipart((b"topic", b"many", b"parts"))
        with self.assertLogs() as cm:
            res = receive_message_from_socket(self.sub_socket)
        self.assertEqual(res, (None, None))
        self.assertIn("invalid number of parts: 3", str(cm.output))

    def test_invalid_json(self):
        self.pub_socket.send_multipart((b"topic", b"ma-_:\"ny"))
        with self.assertLogs() as cm:
            res = receive_message_from_socket(self.sub_socket)
        self.assertEqual(res, (None, None))
        self.assertIn("invalid json", str(cm.output))

    def test_not_a_dict(self):
        payload = b'"asd"'
        self.assertEqual("asd", orjson.loads(payload))

        self.pub_socket.send_multipart((b"topic", payload))
        with self.assertLogs() as cm:
            res = receive_message_from_socket(self.sub_socket)
        self.assertEqual(res, (None, None))
        self.assertIn("not a dict", str(cm.output))



from collections import defaultdict

import logging
import zmq
from pydantic import ValidationError, BaseModel
from typing import Union, Optional, Callable, DefaultDict, List, Dict, TypeVar, Type, Generic

from openmodule.models.base import ZMQMessage


class Listener:
    def __init__(self, message_class: Type[ZMQMessage], filter: Optional[Dict], handler: Callable):
        self.filter = filter
        self.handler = handler
        self.message_class = message_class

    def matches(self, message: Dict):
        if self.filter is None:
            return True
        else:
            return message.items() >= self.filter.items()


EventArgs = TypeVar("EventArgs")


class EventListener(Generic[EventArgs], list):
    log: Optional[logging.Logger]

    def __init__(self, *args, log=None, raise_exceptions=False):
        super().__init__(*args)
        self.raise_exceptions = raise_exceptions
        self.log = log or logging

    def __call__(self, *args: EventArgs):
        for f in self:
            try:
                f(*args)
            except zmq.ContextTerminated:
                raise
            except Exception as e:
                if self.raise_exceptions:
                    raise
                else:
                    self.log.exception(e)


ZMQMessageSub = TypeVar('ZMQMessageSub', bound=ZMQMessage)


class MessageDispatcher:
    def __init__(self, name=None, raise_validation_errors=False, raise_handler_errors=False):
        """
        :param name: optionally name the dispatcher for logging purposes
        :param raise_validation_errors: if true and received messages do not match a validation error is raised,
                                        this is useful in restricive code or testcases
        :param raise_validation_errors: if true and a message handler raises an exception, the exception is raised,
                                        this is useful in restricive code or testcases
        """
        self.name = name
        self.log = logging.getLogger(f"{self.__class__.__name__}({self.name})")
        self.listeners: DefaultDict[bytes, List[Listener]] = defaultdict(list)
        self.raise_validation_errors = raise_validation_errors
        self.raise_handler_errors = raise_handler_errors

    def register_handler(self, topic: Union[bytes, str],
                         message_class: Type[ZMQMessageSub],
                         handler: Callable[[ZMQMessageSub], None], *,
                         filter: Optional[Dict] = None):
        if hasattr(topic, "encode"):
            topic = topic.encode()
        self.listeners[topic].append(Listener(message_class, filter, handler))

    def dispatch(self, topic: bytes, message: Union[Dict, BaseModel]):
        if isinstance(message, BaseModel):
            message = message.dict()

        listeners = self.listeners.get(topic, [])
        for listener in listeners:
            if listener.matches(message):
                self.execute(listener, message)

    def execute(self, listener: Listener, message: Dict):
        try:
            parsed_message = listener.message_class(**message)
        except ValidationError as e:
            if self.raise_validation_errors:
                raise e from None
            else:
                self.log.exception("Invalid message received")
        else:
            try:
                listener.handler(parsed_message)
            except zmq.ContextTerminated:
                raise
            except Exception as e:
                if self.raise_handler_errors:
                    raise e from None
                else:
                    self.log.exception("Error in message handler")


class SubscribingMessageDispatcher(MessageDispatcher):
    def __init__(self, subscribe: Callable[[bytes], None], name=None):
        super().__init__(name=name)
        self.subscribe = subscribe

    def register_handler(self, topic: Union[bytes, str],
                         message_class: Type[ZMQMessageSub],
                         handler: Callable[[ZMQMessageSub], None], *,
                         filter: Optional[Dict] = None):
        self.subscribe(topic)
        return super().register_handler(topic, message_class, handler, filter=filter)



import logging
import threading
from typing import List, Dict, Optional

from openmodule.models.alert import AlertHandleType, AlertMessage, AlertStatus


class AlertHandler(object):
    _alerts: List[Dict]

    def __init__(self, core: 'OpenModuleCore'):
        self._alerts = []
        self._alerts_lock = threading.RLock()
        self.core = core
        self.log = logging.getLogger(self.__class__.__name__)

    def _send_message(self, alert_id, status_dict):
        alert = self._alerts[alert_id]
        if alert["alert_dict"]["handle_type"] == AlertHandleType.state and status_dict.get("value") is None:
            raise ValueError("value must not be none for alerts with handle_type=state")

        status_dict["alert_meta"].update(dict(sender=self.core.config.NAME))
        old_status = alert["alert_dict"].get("status")
        alert["alert_dict"].update(status_dict)
        if alert["always_send_alert"] or alert["alert_dict"]["status"] != old_status:
            message = AlertMessage(
                name=self.core.config.NAME,
                type="alert",
                **alert["alert_dict"]
            )

            package = message.package if message.package != self.core.config.NAME else ""
            source = message.source
            tmp = f"{' from ' if package or source else ''}{package}{'/' if package and source else ''}{source}:"
            self.log.info(f"Send alert{tmp} {message.alert_type} - {message.status}")

            self.core.publish(message, b"alert")
            return True
        return False

    def get_or_add_alert_id(self, source: str, alert_type: str, handle_type: AlertHandleType,
                            package_name: Optional[str] = None, version: Optional[str] = None,
                            always_send_alert: Optional[bool] = None):
        """Define an alert with its name parameters

        Args:
            source (str): The source of the error (i.e. which pin, camera)
            alert_type (str): Name of the alert all lower_case with '_' (i.e. high_time_check)
            handle_type (str): How the error should be handled on the server (state, state_change, count)
            package_name (str) (default=config.NAME): You send the error for this package
            version (str) (default=config.VERSION): Current version of the package
            always_send_alert (bool) (default=False): Send every alert or only when there is a change in the status

        Returns:
            alert_id of the saved alert
        """

        assert "-" not in alert_type, "please do not use '-' in your alert type names"

        alert_dict = dict(source=source, alert_type=alert_type, handle_type=handle_type)
        alert_dict["package"] = package_name or self.core.config.NAME

        with self._alerts_lock:
            for alert in self._alerts:
                if alert_dict.items() <= alert["alert_dict"].items():
                    if always_send_alert is not None:
                        alert["always_send_alert"] = always_send_alert
                    if version is not None:
                        alert["alert_dict"]["version"] = version
                    return alert["id"]

            v = version if version is not None else self.core.config.VERSION
            asa = always_send_alert if always_send_alert is not None else False
            alert = {"alert_dict": alert_dict, "id": len(self._alerts), "always_send_alert": asa, "version": v}
            self._alerts.append(alert)
            return alert["id"]

    def send_with_alert_id(self, alert_id: int, status: AlertStatus, meta: Dict, value: Optional[float] = None):
        """ Send a status message on the given alert with the corresponding meta/kwarg arguments
            Returns True if alert was send else False"""
        assert status in [AlertStatus.error, AlertStatus.ok, AlertStatus.offline]

        status_dict = {
            "status": status,
            "alert_meta": (meta or {}),
        }
        if value is not None:
            status_dict["value"] = value

        return self._send_message(alert_id, status_dict)

    def send(self, source: str, alert_type: str, handle_type: AlertHandleType,
             package_name: Optional[str] = None, version: Optional[str] = None,
             always_send_alert: Optional[bool] = None, status: AlertStatus = AlertStatus.error,
             meta: Optional[Dict] = None, value: Optional[float] = None):

        """Sends an alert to the device serve

        Args:
            source (str): The source of the error (i.e. which pin, camera)
            alert_type (str): Name of the alert all lower_case with '_' (i.e. high_time_check, health_hw_compute_nuc)
            handle_type (str): How the error should be handled on the server (state, state_change, count)
            status (Status/bool): The status you want to send(Status.OK/True, Status.ERROR/False, Status.OFFLINE)
            package_name (str) optional: You can send errors for other packages
            version (str) (default=config.VERSION): Current version of the package
            always_send_alert (bool): Flag to declare if you want to send every alert or only every change in the status
            meta (dict): Meta information
            value (float): value of the error metric, only applicable if handle_type=state

        Returns:
            True if alert was send else False
        """
        assert "-" not in alert_type, "please do not use - in your alert type names"

        alert_id = self.get_or_add_alert_id(source, alert_type, handle_type, package_name, version, always_send_alert)
        return self.send_with_alert_id(alert_id, status, meta, value)



import logging


def init_logging(core):
    assert hasattr(core.config, "LOG_LEVEL"), (
        "LOG_LEVLE setting not found in your config. In order to use logging please add \n"
        "> LOG_LEVEL = config.log_level()\n"
        "to your config.py"
    )
    logging.basicConfig(level=core.config.LOG_LEVEL)
    logging.captureWarnings(True)



# coding=utf-8
import codecs
import os
import random
import shutil
import sys
import tempfile
import unittest

from openmodule.utils.matching import PlateYamlSelector

plates = ["G ARIVO 1", "BÖ SE-11", "GÜ-NTH ER 123", "OQ:123 AB",
          "SLDFKJWPEOÄÜSDGIEFH .:S.F: WE:. OWEKRQÖ.Ö: .WEIJFA.DF", "ASDFÖÜĆẐ"]
legacy_results = ["GARIV01", "B0SE11", "GUNTHER123", "00123AB", "SLDFKJWPE0AUSDGIEFHSFWE0WEKR000WEIJFADF", "ASDF0UCZ"]
a10_results = ["G:ARIVO1", "BO:SE11", "GU:NTHER123", "OO:123AB",
               "SLDFKJWPEOAUSDGIEFH:SFWEOWEKROOOWEIJFADF", "ASDFOUCZ"]
a20_results = ["G ARIVO1", "BO SE11", "GU-NTHER123", "OO:123AB",
               "SLDFKJWPEOAUSDGIEFH SFWEOWEKROOOWEIJFADF", "ASDFOUCZ"]
d10_results = ["GARIVO1", "BÖSE11", "GÜNTHER123", "OQ123AB", "SLDFKJWPEOÄÜSDGIEFHSFWEOWEKRQÖÖWEIJFADF", "ASDFÖÜCZ"]
d20_results = ["G:ARIVO:1", "BÖ:SE:11", "GÜ:NTH:ER:123", "OQ:123:AB",
               "SLDFKJWPEOÄÜSDGIEFH:::S:F::WE:::OWEKRQÖ:Ö:::WEIJFA:DF", "ASDFÖÜCZ"]
d30_results = ["G ARIVO 1", "BÖ SE-11", "GÜ-NTH ER 123", "OQ:123 AB",
               "SLDFKJWPEOÄÜSDGIEFH .:S.F: WE:. OWEKRQÖ.Ö: .WEIJFA.DF", "ASDFÖÜCZ"]
t10_results = ["GARIVO1", "BÖSE11", "GÜNTHER123", "OQ123AB",
               "SLDFKJWPEOÄÜSDGIEFHSFWEOWEKRQÖÖWEIJFADF", "ASDFÖÜCZ"]
t20_results = ["G ARIVO1", "BÖ SE11", "GÜ-NTHER123", "OQ:123AB",
               "SLDFKJWPEOÄÜSDGIEFH SFWEOWEKRQÖÖWEIJFADF", "ASDFÖÜCZ"]
t30_results = ["GARIVO 1", "BÖSE-11", "GÜNTHER 123", "OQ123 AB",
               "SLDFKJWPEOÄÜSDGIEFHSFWEOWEKRQÖÖWEIJFA.DF", "ASDFÖÜCZ"]
t40_results = ["G ARIVO 1", "BÖ SE-11", "GÜ-NTH ER 123", "OQ:123 AB",
               "SLDFKJWPEOÄÜSDGIEFH .:S.F: WE:. OWEKRQÖ.Ö: .WEIJFA.DF", "ASDFÖÜCZ"]
all_results = [[["A", 10], a10_results], [["A", 20], a20_results], [["DEFAULT", 10], d10_results],
               [["DEFAULT", 20], d20_results], [["DEFAULT", 30], d30_results], [["TEST", 10], t10_results],
               [["TEST", 20], t20_results], [["TEST", 30], t30_results], [["TEST", 40], t40_results]]


def gen_plate(plates, countries):
    co = []
    for c in countries:
        co.append({"code": c, "confidence": 0.5})
    pl = []
    for p in plates:
        pl.append({"plate": p, "confidence": 0.1})

    return {
        "plate": {"plate": plates[0], "confidence": 0.1},
        "country": co,
        "alternatives": pl
    }


def prepare_db(plates, default_country_order=[], dco_w_legacy=False, caps=[]):
    pys = PlateYamlSelector(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"), "",
                            default_country_order, dco_w_legacy)
    db = {}
    for i, p in enumerate(plates):
        pys.insert_into_db_and_create_record(db, p[0], "SRB" if p[1] == "TEST" else p[1], p[1], p[2], i)

    return db, default_country_order, dco_w_legacy


plates_mo = [gen_plate(["G ÄRIVQ-0"], ["F", "A", "SRB", "OTHER"]),
             gen_plate(["G ÄRIVQ-0"], ["A", "F", "SRB", "OTHER"]),
             gen_plate(["G ÄRIVQ-0"], ["SRB", "A", "F", "OTHER"]),
             gen_plate(["G ÄRIVQ-0"], ["D", "OTHER"]),
             gen_plate(["G ÄRIVQ-0"], ["OTHER"]),
             gen_plate(["GARIV00"], ["F", "A", "SRB", "OTHER"]),
             gen_plate(["GARIV00"], ["A", "F", "SRB", "OTHER"]),
             gen_plate(["GARIV00"], ["SRB", "A", "F", "OTHER"]),
             gen_plate(["GARIV00"], ["D", "OTHER"]),
             gen_plate(["GARIV00"], ["OTHER"]),
             gen_plate(["G ARIVO0"], ["F", "A", "SRB", "OTHER"]),
             gen_plate(["G ARIVO0"], ["A", "F", "SRB", "OTHER"]),
             gen_plate(["G ARIVO0"], ["SRB", "A", "F", "OTHER"]),
             gen_plate(["G ARIVO0"], ["D", "OTHER"]),
             gen_plate(["G ARIVO0"], ["OTHER"]),
             gen_plate(["G:ARIV00"], ["F", "A", "SRB", "OTHER"]),
             gen_plate(["G:ARIV00"], ["A", "F", "SRB", "OTHER"]),
             gen_plate(["G:ARIV00"], ["SRB", "A", "F", "OTHER"]),
             gen_plate(["G:ARIV00"], ["D", "OTHER"]),
             gen_plate(["G:ARIV00"], ["OTHER"]),
             gen_plate(["X-YZ"], ["F", "A", "SRB", "OTHER"]),
             gen_plate(["X-YZ"], ["A", "F", "SRB", "OTHER"]),
             gen_plate(["X-YZ"], ["SRB", "A", "F", "OTHER"]),
             gen_plate(["X-YZ"], ["D", "OTHER"]),
             gen_plate(["X-YZ"], ["OTHER"]),
             gen_plate(["G-ÄRIVQ-0"], ["SRB"]),
             gen_plate(["G-ÄRIVQ-0"], ["CH"])
             ]

results_mo = [[3, 0, 2, 1, 2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 4, 5, 5, 6, 4, -1, -1],
              [3, 0, 2, 1, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 4, 5, 5, 6, 5, -1, -1],
              [3, 0, 2, 1, 1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 4, 5, 5, 6, 4, -1, -1],
              [3, 0, 2, 1, 1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 4, 5, 5, 6, 6, -1, -1],
              [3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 4, 5, 5, 6, 4, 2, -1],
              [3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 4, 5, 5, 6, 5, 2, -1],
              [3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 4, 5, 5, 6, 4, 2, -1],
              [3, 0, 2, 1, 2, 3, 0, 2, 1, 2, 3, 0, 2, 1, 2, 3, 0, 2, 1, 2, 4, 5, 5, 6, 6, 2, -1],
              [3, 0, 2, 1, 1, 3, 0, 2, 1, 1, 3, 0, 2, 1, 1, 3, 0, 2, 1, 1, 4, 5, 5, 6, 6, 2, -1],
              [3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2, 1, 0, 4, 5, 5, 6, 4, 2, -1]]

dbs = [
    prepare_db([("G ÄRIVQ-0", "A", 100), ("G ÄRIVQ-0", "D", 100), ("G ÄRIVQ-0", "TEST", 100), ("G ÄRIVQ-0", "F", 100),
                ("X-YZ", "F", 100), ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
               []),
    prepare_db([("G ÄRIVQ-0", "A", 100), ("G ÄRIVQ-0", "D", 100), ("G ÄRIVQ-0", "TEST", 100), ("G ÄRIVQ-0", "F", 100),
                ("X-YZ", "F", 100), ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
               ['A']),
    prepare_db([("G ÄRIVQ-0", "A", 100), ("G ÄRIVQ-0", "D", 100), ("G ÄRIVQ-0", "TEST", 100), ("G ÄRIVQ-0", "F", 100),
                ("X-YZ", "F", 100), ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
               ['D']),
    prepare_db([("G ÄRIVQ-0", "A", 100), ("G ÄRIVQ-0", "D", 100), ("G ÄRIVQ-0", "TEST", 100), ("G ÄRIVQ-0", "F", 100),
                ("X-YZ", "F", 100), ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
               ['D'], dco_w_legacy=True),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        []),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        ['A']),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        ['SRB', "D"]),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        ['SRB', "D"], dco_w_legacy=True),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        ['D'], dco_w_legacy=True),
    prepare_db(
        [("GARIV00", "A", 0), ("GARIV00", "D", 0), ("GARIV00", "TEST", 0), ("GARIV00", "F", 0), ("X-YZ", "F", 100),
         ("X-YZ", "A", 10), ("X-YZ", "D", 0)],
        ['X'], dco_w_legacy=True)
]

plates_alt = [gen_plate(["X-YZ", "X YZ"], ["F", "OTHER"]),
              gen_plate(["X-YZ", "X YZ"], ["OTHER"]),
              gen_plate(["X-YZ", "XYZ", "X YZ"], ["OTHER"]),
              gen_plate(["XYZA", "XYZ:"], ["D"]),
              gen_plate(["XYZA", "XYZ:"], ["CH"])]
results_alt = [[0, 0, 0, 2, -1], [2, 2, 2, 2, -1], [3, 1, 2, -1, -1]]
dbs_alt = [
    prepare_db([("X:YZ", "A", 10), ("X YZ", "TEST", 40), ("XYZ", "D", 0), ("X YZ", "F", 30)], []),
    prepare_db([("X YZ", "A", 20), ("X YZ", "TEST", 40), ("XYZ", "D", 0), ("X YZ", "F", 30)], []),
    prepare_db([("X YZ", "A", 20), ("X YZ", "TEST", 40), ("XYZ", "D", 30), ("X YZ", "F", 30)], [])
]


class MatchingTest(unittest.TestCase):
    def do_stuff(self, results, scheme, version):
        pys = PlateYamlSelector(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"),
                                "")
        for p, r in zip(plates, results):
            converted = pys.convert(p, scheme, version)
            success = r == converted
            if not success:  # pragma: no cover
                print(scheme, version)
                print(p, "->", converted)
            self.assertTrue(success)

    def test_legacy_convert(self):
        self.do_stuff(legacy_results, "LEGACY", 0)

    def test_selection(self):
        pys = PlateYamlSelector(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"),
                                "")
        inputs = [["LEGACY", 0], ["LEGACY", 1], ["DEFAULT", 0], ["DEFAULT", 1], ["DEFAULT", 10], ["DEFAULT", 11],
                  ["DEFAULT", 20], ["DEFAULT", 25], ["DEFAULT", 35], ["A", 5], ["A", 10], ["A", 15], ["A", 20],
                  ["A", 25], ["TEST", 5], ["TEST", 55], ["ASD", 0], ["aeg", 1], ["whbns", 10],
                  ["sh", 11], ["af", 20], ["hs", 25], ["jsd", 35]]
        results = [["LEGACY", 0], ["LEGACY", 0], ["LEGACY", 0], ["LEGACY", 0], ["DEFAULT", 10], ["DEFAULT", 10],
                   ["DEFAULT", 20], ["DEFAULT", 20], ["DEFAULT", 30], ["LEGACY", 0], ["A", 10], ["A", 10], ["A", 20],
                   ["A", 20], ["LEGACY", 0], ["TEST", 40], ["LEGACY", 0], ["LEGACY", 0], ["DEFAULT", 10],
                   ["DEFAULT", 10], ["DEFAULT", 20], ["DEFAULT", 20], ["DEFAULT", 30]]
        for i, r in zip(inputs, results):
            yaml = pys.get_best_yaml(*i)
            self.assertEqual(r, [yaml.scheme, yaml.version])

    def test_various_conversions(self):
        for res in all_results:
            self.do_stuff(res[1], res[0][0], res[0][1])

    def test_matching_order_no_alternatives(self):
        for i, db1 in enumerate(dbs):
            db = db1[0]
            pys = PlateYamlSelector(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"), "", db1[1],
                db1[2])
            for j, plate in enumerate(plates_mo):
                res = pys.match_plate(plate, db)
                res = None if len(res[0]) == 0 else res[0][0]
                sys.stdout.flush()
                success = results_mo[i][j] == (-1 if res is None else res["id"])
                if not success:  # pragma: no cover
                    print("database", i)
                    print("plate", j, ":", plate["plate"], "|", plate["country"][0]["code"], "|", results_mo[i][j],
                          "->")
                    if res is not None:
                        print(res["plate"], "|", res["plate_scheme"], "|", res["plate_scheme_version"], "|",
                              res["id"], "|", res["used_scheme"], "|", res["used_version"], "|", res["exactitude"])
                    else:
                        print()
                self.assertTrue(success)

    def test_matching_order_with_alternatives(self):
        for i, db1 in enumerate(dbs_alt):
            db = db1[0]
            pys = PlateYamlSelector(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"), "", db1[1],
                db1[2])
            for j, plate in enumerate(plates_alt):
                res = pys.match_plate(plate, db, return_only_best=False)
                res = None if len(res[0]) == 0 else res[0][0]
                sys.stdout.flush()
                success = results_alt[i][j] == (-1 if res is None else res["id"])
                if not success:  # pragma: no cover
                    print("database", i)
                    print("plate", j, ":", plate["plate"], "|", plate["country"][0]["code"], "|", results_alt[i][j],
                          "->")
                    if res is not None:
                        print(res["plate"], "|", res["plate_scheme"], "|", res["plate_scheme_version"], "|",
                              res["id"], "|", res["used_scheme"], "|", res["used_version"], "|", res["exactitude"])
                    else:
                        print()
                self.assertTrue(success)

    def test_matching_order_ed1(self):
        plates_ed1 = [gen_plate(["X-YZ"], ["F", "OTHER"]), gen_plate(["X-YZ"], ["OTHER"]), gen_plate(["XYZA"], ["D"]),
                      gen_plate(["XYZA"], ["CH"])]
        results_ed1 = [[0, 0, 2, -1], [2, 2, 2, -1], [2, 1, -1, -1], [2, 0, -1, -1]]
        dbs_ed1 = [
            prepare_db([("X:YZ", "A", 10), ("X YZ", "TEST", 40), ("XYZ", "D", 0), ("X YZ", "F", 30)], []),
            prepare_db([("X YZ", "A", 20), ("X YZ", "TEST", 40), ("XYZ", "D", 0), ("X YZ", "F", 30)], []),
            prepare_db([("X YZ", "A", 20), ("X YZ", "TEST", 40), ("X YZ", "F", 30)], []),
            prepare_db([("X YZ", "A", 20), ("X YZ", "TEST", 40), ("X YZ", "F", 30)], ["A"])
        ]

        for i, db1 in enumerate(dbs_ed1):
            db = db1[0]
            pys = PlateYamlSelector(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"), "", db1[1],
                db1[2])
            for j, plate in enumerate(plates_ed1):
                res = pys.match_plate(plate, db, False, 1)
                res = None if len(res[0]) == 0 else res[0][0]
                sys.stdout.flush()
                success = results_ed1[i][j] == (-1 if res is None else res["id"])
                if not success:  # pragma: no cover
                    print("database", i)
                    print("plate", j, ":", plate["plate"], "|", plate["country"][0]["code"], "|", results_ed1[i][j],
                          "->")
                    if res is not None:
                        print(res["plate"], "|", res["plate_scheme"], "|", res["plate_scheme_version"], "|",
                              res["id"], "|", res["used_scheme"], "|", res["used_version"], "|", res["exactitude"])
                    else:
                        print()
                self.assertTrue(success)

    def test_random(self):
        dbs = []
        for i in range(200):
            pdbs = []
            for i in range(100):
                pdbs.append((''.join(
                    random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. ") for _ in range(random.randint(5, 10))),
                             ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in
                                     range(random.randint(1, 5))), random.randint(0, 10) * 10))
            dbs.append(prepare_db(pdbs, [
                ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2))),
                ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2))),
                ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2)))],
                                  random.randint(0, 1)))
        ps = []
        for i in range(1000):
            ps.append(gen_plate([''.join(
                random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. ") for _ in range(random.randint(5, 10))),
                ''.join(
                    random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. ") for _ in range(random.randint(5, 10))),
                ''.join(
                    random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. ") for _ in range(random.randint(5, 10))),
                ''.join(
                    random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. ") for _ in
                    range(random.randint(5, 10)))],
                [''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2))),
                 ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2))),
                 ''.join(random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ") for _ in range(random.randint(1, 2))),
                 ]))
        for db in dbs:
            pys = PlateYamlSelector(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"), "", db[1], db[2])
            for p in ps:
                pys.match_plate(p, db[0], random.randint(0, 1), random.randint(0, 1), return_only_best=False)

    def test_no_yamls(self):
        pys = PlateYamlSelector("/")
        self.assertEqual("GARIV01", pys.convert("G ARIVO 1", "A", 10))

    def test_reload(self):
        tempdir = tempfile.mktemp()
        os.makedirs(tempdir, exist_ok=True)
        try:
            revision_path = os.path.join(tempdir, "revision")
            default_10_path = os.path.join(tempdir, "DEFAULT-10.yml")

            with open(revision_path, 'w') as f:
                f.write("1")
            pys = PlateYamlSelector(tempdir, revision_path)
            self.assertEqual("GARIV01", pys.convert("G ARIVO 1", "A", 10))
            with open(default_10_path, "w") as f:
                f.write("""
    allowed: "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ:-. "
    kept_separators: ALL
    exactitude: 1.3
    """)

            # revision did not change, no reload is happening
            pys.reload_yamls()
            self.assertEqual("GARIV01", pys.convert("G ARIVO 1", "A", 10))

            # change revision to trigger reload
            with open(revision_path, 'w') as f:
                f.write("2")
            pys.reload_yamls()
            self.assertEqual("G ARIVO 1", pys.convert("G ARIVO 1", "A", 10))
        finally:
            shutil.rmtree(tempdir)

    def test_performance(self):
        def print_time(name):
            diff = time.time() - start
            print(name, f"{diff:.02f} for {count} = {(count / diff):.02f} per second")

        import time
        ps = []
        count = 100000
        start = time.time()
        initial_start = start
        for i in range(count):
            ps.append((''.join(
                random.choice("0123456789ABCDEFGHIJKLMNOPQRSTUVWX-YZÄÖÜ:-. ") for _ in range(random.randint(5, 10))),
                       random.choice(["A", "D", "TEST", "F"]), 100))
        print_time("Random Plates")

        start = time.time()
        db = prepare_db(ps, ['A', 'D'], False)
        print_time("Prepare DB")

        pys = PlateYamlSelector(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/utils_matching"),
                                "", db[1], db[2])

        start = time.time()
        pys.match_plate(plates_mo[3], db[0])
        print_time("Match plate")

        start = initial_start
        print_time("Total time")



import logging
import os
import warnings
from socket import gethostname
from typing import Optional, Type, TypeVar

import builtins
import yaml
from pydantic import ValidationError
from pydantic.main import BaseModel


def _save_cast(obj, key, to_type, default):
    try:
        return to_type(obj[key])
    except (ValueError, TypeError, KeyError):
        return default


def int(key: str, default: int = 0) -> int:
    assert isinstance(default, (builtins.int,)), "default argument must be of type int"
    return _save_cast(os.environ, key, builtins.int, builtins.int(default))


def float(key: str, default: float = 0.0) -> float:
    assert isinstance(default, (builtins.int, builtins.float)), "default argument must be of type int or float"
    return _save_cast(os.environ, key, builtins.float, builtins.float(default))


def string(key: str, default: str = "") -> str:
    assert isinstance(default, str), "default argument must be of type string"
    res = os.environ.get(key, default)
    invalid_chars = "\"' \r\n\t"
    res = res.strip(invalid_chars)
    return res


def bool(key: str, default: bool = False) -> bool:
    assert isinstance(default, builtins.bool), "default argument must be of type bool"
    val = get(key) or ""
    val = val.upper().strip('\'" .,;-_')
    if not val:
        return default
    return val in ["T", "TRUE", "1", "Y", "YES", "J", "JA", "ON"]


def get(key: str, default: Optional[str] = "") -> str:
    return os.environ.get(key, default)


_inside_docker_cached = None


def _inside_docker():
    global _inside_docker_cached
    if _inside_docker_cached is not None:
        return _inside_docker_cached
    else:
        if os.path.exists('/.dockerenv'):
            _inside_docker_cached = True
            return _inside_docker_cached

        path = '/proc/self/cgroup'
        if os.path.isfile(path):
            with open(path, "r") as f:
                _inside_docker_cached = any('docker' in line for line in f)
        else:
            _inside_docker_cached = False

        return _inside_docker_cached


def debug():
    if "DEBUG" in os.environ:
        return bool("DEBUG", default=False)
    elif testing():
        return False
    else:
        return version() == "unknown" and not _inside_docker()


def log_level() -> int:
    log_levels = {
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "VERBOSE": logging.DEBUG,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "ERR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
        "CRIT": logging.CRITICAL,
        "FATAL": logging.FATAL
    }
    default_log_level = logging.DEBUG if debug() else logging.INFO
    return log_levels.get(os.environ.get("LOG_LEVEL", "").upper(), default_log_level)


def version() -> str:
    version = string("VERSION", "")
    if not version:
        if os.path.isfile("VERSION"):
            with open("VERSION", "r") as f:
                version = f.read()
    if not version:
        version = "unknown"
    return version.strip("\r\n\t ")


def resource() -> str:
    resource = string("AUTH_RESOURCE", "")
    if not resource:
        resource = gethostname()
    return resource


def broker_pub(default: str = "tcp://127.0.0.1:10200") -> str:
    broker_host = string("BROKER_HOST", "")
    broker_pub_port = int("BROKER_PUB_PORT", 0)
    broker_pub = string("BROKER_PUB", "")

    if broker_pub:
        return broker_pub
    elif broker_host and broker_pub_port:
        # use legacy configs
        return "tcp://{}:{}".format(broker_host, broker_pub_port)
    else:
        return default


def broker_sub(default: str = "tcp://127.0.0.1:10100") -> str:
    broker_host = string("BROKER_HOST", "")
    broker_sub_port = int("BROKER_SUB_PORT", 0)
    broker_sub = string("BROKER_SUB", "")

    if broker_sub:
        return broker_sub
    elif broker_host and broker_sub_port:
        # use legacy configs
        return "tcp://{}:{}".format(broker_host, broker_sub_port)
    else:
        return default


def validate_config_module(config):
    required_keys = ["NAME", "RESOURCE", "VERSION", "DEBUG", "TESTING"]
    for x in required_keys:
        assert hasattr(config, x), f"the config module requires the variable {x}"

    if hasattr(config, "BROKER_PUB_PORT") or hasattr(config, "BROKER_SUB_PORT"):
        warnings.warn(
            "BROKER_(P/S)UB_PORT and BROKER_HOST are to be replaced with BROKER_(P/S)UB. In order to \n"
            "remain backwards compatible, you can use BROKER_PUB = config.broker_pub() which correctly \n"
            "interprets the deprecated environment variables",
            DeprecationWarning
        )


class SafeLoaderIgnoreUnknown(yaml.SafeLoader):
    def ignore_unknown(self, node):
        return None


SafeLoaderIgnoreUnknown.add_constructor(None, SafeLoaderIgnoreUnknown.ignore_unknown)

YamlType = TypeVar("YamlType", bound=BaseModel)


def yaml(model: Type[YamlType], path: str = None) -> YamlType:
    yaml_path = path or string("CONFIG_YAML", "")
    if not yaml_path:
        if debug():
            yaml_path = "../settings/default-debug.yml"
        elif testing():
            yaml_path = "../settings/default-testing.yml"
        else:
            yaml_path = "/data/config.yml"

    try:
        if os.path.exists(yaml_path):
            import yaml
            with open(yaml_path, "r") as f:
                a = yaml.load(f, SafeLoaderIgnoreUnknown)
                return model.parse_obj(a)
        else:
            return model()
    except ValidationError as e:
        logging.exception("error during config yaml loading, something is wrong with the configuration")
        raise e from None


def testing():
    return bool("TESTING", False)


def database_folder():
    path = string("DATABASE_FOLDER", f"../sqlite/")
    if testing():
        path += "test/"
    elif debug():
        path += "debug/"
    return path



import json
import time
from datetime import datetime
from textwrap import dedent
from typing import List
from unittest import TestCase

from openmodule.models.backend import CountMessage, AccessRequest, Access, Category, MediumAccesses
from openmodule.models.base import Gateway
from openmodule.rpc.server import RPCServer
from openmodule.utils.backend import Backend
from openmodule_test.backend import TestBackend, BackendTestMixin
from openmodule_test.eventlistener import MockEvent
from openmodule_test.rpc import RPCServerTestMixin


class ModelTest(TestCase):
    def test_count_message(self):
        messages = dedent("""  [{  "real": true,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "timestamp": 1536154434.152993,
                    "count": 1,
                    "gateway": {
                        "gate": "gate1",
                        "direction": "in"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "access_data": {
                        "on_presence_loop": true,
                        "has_presence_loop": false,
                        "decision": "accepted"
                    },
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "dc4e4af3-5ff4-4258-a310-089701ef37e8" }, 
                    { "real": false,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "timestamp": 1536154434.19146,
                    "count": 0,
                    "gateway": {
                        "gate": "automatically_deleted",
                        "direction": "out"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "dc4e4af3-5ff4-4258-a310-089701ef37e8"},
                     {  "real": true,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "previous_transaction_id": ["dc4e4af3-5ff4-4258-a310-089701ef37e8"],
                    "timestamp": 1536154434.192287,
                    "count": 1,
                    "gateway": {
                        "gate": "gate1",
                        "direction": "in"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "error": "double_entry",
                    "access_data": {
                        "on_presence_loop": true,
                        "has_presence_loop": false,
                        "decision": "accepted"
                    },
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "0337597c-f125-44e0-b38f-a6537e7d38d1" }]
                """).strip()
        messages = json.loads(messages)
        for message in messages:
            model = CountMessage(**message)
            self.assertTrue(bool(model.transaction_id))

    def test_access_recurrence(self):
        kwargs = dict(user="asdf", category="permanent-employee", start=3,
                      recurrence="DTSTART:20180702T123000\nFREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR")
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("set duration", str(e.exception))

        kwargs["duration"] = 0
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("set duration", str(e.exception))

        kwargs["duration"] = 1000
        access = Access(**kwargs)
        self.assertEqual(1000, access.duration)

        kwargs["recurrence"] = "asdf"
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("recurrence is not valid", str(e.exception))

    def test_access_with_empty_string_recurrence_is_valid(self):
        access_as_dict = {
            "user": "user",
            "category": "filler-visitor-button",
            "start": time.time(),
            "recurrence": "",
        }
        try:
            Access(**access_as_dict)
        except ValueError:
            self.fail("access is invalid with empty string recurrence")

    def test_category(self):
        with self.assertRaises(Exception) as e:
            Category("test_exception")
        self.assertIn("not a valid category", str(e.exception).lower())

        with self.assertRaises(Exception) as e:
            Access(user="asdf", category="permanent_employee", start=3,
                   recurrence="DTSTART:20180702T123000\nFREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR")
        self.assertIn("not a valid enumeration", str(e.exception))


class NotImplementedTestCase(BackendTestMixin, TestCase):
    backend_class = Backend
    config_kwargs = dict(USER_PREFIX="backend")

    # noinspection PyTypeChecker
    def test_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.backend.check_in(None)
        with self.assertRaises(NotImplementedError):
            self.backend.check_out(None)
        with self.assertRaises(NotImplementedError):
            self.backend.check_access(None)


class FunctionTest(BackendTestMixin, TestCase):
    config_kwargs = dict(USER_PREFIX="backend")
    backend_class = TestBackend

    def test_check_in(self):
        transaction_id = self.check_in()
        message = self.messages_by_transaction.get(transaction_id)
        self.assertEqual(None, message.error)

        transaction_id = self.check_in_double_entry(transaction_id)
        message = self.messages_by_transaction.get(transaction_id)
        self.assertEqual("double_entry", message.error)
        self.assertNotEqual(None, message.previous_transaction_id)

        with self.assertRaises(Exception) as e:
            self.check_in(gate="error")
        self.assertIn("test_exception", str(e.exception))

    def check_in_out(self):
        transaction_id = self.check_in()
        self.assertNotEqual(None, self.messages_by_transaction.get(transaction_id))

        self.check_out(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))

    def test_check_out(self):
        with self.assertRaises(Exception) as e:
            self.check_out("asdf")
        self.assertIn("asdf", str(e.exception))

        transaction_id = self.check_in(send=False)
        self.messages_by_transaction[transaction_id].gateway.gate = "error"
        with self.assertRaises(Exception) as e:
            self.check_out(transaction_id)
        self.assertIn("test_exception", str(e.exception))

        transaction_id = self.check_in(send=False)
        message = self.check_out(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual(None, message.error)

        # category_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_category_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("category_changed", message.error)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_category_changed(transaction_id, category="permanent-employee")
        self.assertIn("category needs to change", str(e.exception))

        # medium_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_medium_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("medium_type_changed", message.error)
        self.assertNotEqual(None, message.previous_id)
        self.assertNotEqual(None, message.previous_medium_type)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_medium_changed(transaction_id, medium="lpr")
        self.assertIn("medium needs to change", str(e.exception))

        # medium_id_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_medium_id_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("medium_id_changed", message.error)
        self.assertNotEqual(None, message.previous_id)

        transaction_id = self.check_in(send=False)
        medium_id = self.messages_by_transaction[transaction_id].id
        with self.assertRaises(Exception) as e:
            self.check_out_medium_id_changed(transaction_id, medium_id=medium_id)
        self.assertIn("medium_id needs to change", str(e.exception))

        # user_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_user_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("user_changed", message.error)
        self.assertNotEqual(None, message.previous_user)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_user_changed(transaction_id, user="api-backend_user")
        self.assertIn("user needs to change", str(e.exception))

    def test_access(self):
        result = self.check_auth()
        self.assertEqual(True, result.success)
        self.assertNotEqual([], result.medium.accesses)

        result = self.check_auth(gate="empty")
        self.assertEqual(True, result.success)
        self.assertEqual([], result.medium.accesses)

        result = self.check_auth(gate="error")
        self.assertEqual(False, result.success)


class RpcTest(BackendTestMixin, RPCServerTestMixin, TestCase):
    rpc_channels = ["backend"]
    topics = [b"backend", b"count", b"healthz"]
    config_kwargs = dict(USER_PREFIX="backend")
    backend_class = TestBackend

    def setUp(self):
        super().setUp()
        self.server = RPCServer(config=self.zmq_config(), context=self.zmq_context())
        self.server.run_as_thread()
        self.backend.register_rpcs(self.server)
        self.wait_for_rpc_server(self.server)

    def tearDown(self):
        self.server.shutdown()
        super().tearDown()

    def test_access_rpc(self):
        # only test if rpc is passed on to function, function is tested in previous TestCase
        gateway = Gateway(gate="gate", direction="in")
        request = AccessRequest(gateway=gateway, name=self.core.config.NAME, medium_type="lpr", id="medium_id")
        response = self.rpc("backend", "auth", request=request.dict())
        self.assertRPCSuccess(response)
        self.assertIn(request.id, self.backend.accessed)

    def send_count_message(self, gate="gate", direction="in"):
        transaction_id = self.create_count_message(direction=direction, gate=gate)
        self.zmq_client.send("count", **self.messages_by_transaction[transaction_id].dict())

    def test_registration(self):
        # auto register on start
        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register", message["type"])

        self.zmq_client.send("backend", type="register_request")
        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register_request", message["type"])

        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register", message["type"])

    def test_check_in(self):
        with self.assertLogs() as cm:
            self.send_count_message()
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("successful", cm.output[0])

        MockEvent.reset_all_call_counts()
        with self.assertLogs() as cm:
            self.send_count_message(gate="error")
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("failed", cm.output[0])

    def test_check_out(self):
        with self.assertLogs() as cm:
            self.send_count_message()
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("successful", cm.output[0])

        MockEvent.reset_all_call_counts()
        with self.assertLogs() as cm:
            self.send_count_message(gate="error", direction="out")
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("failed", cm.output[0])

        MockEvent.reset_all_call_counts()
        with self.assertLogs() as cm:
            self.send_count_message(direction="out")
            self.backend.wait_for_message_process()
        self.assertEqual(0, self.backend.checked_in)
        self.assertIn("successful", cm.output[0])


class TestBackendCheckAccessWithDatabase(Backend):
    __test__ = False  # otherwise pytest thinks this is a testcase

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message_processed = MockEvent()

    database: List[MediumAccesses] = []

    def check_access(self, request: AccessRequest):
        medium: MediumAccesses = next((m for m in self.database if m.id == request.id))
        if medium.type == request.medium_type:
            return medium.accesses
        return []


class AccessDatetimesAreConvertedToUnixTimestamps(BackendTestMixin, RPCServerTestMixin, TestCase):
    rpc_channels = ["backend"]
    topics = [b"backend", b"healthz"]
    config_kwargs = dict(USER_PREFIX="test")
    backend_class = TestBackendCheckAccessWithDatabase

    def setUp(self):
        super().setUp()
        self.server = RPCServer(config=self.zmq_config(), context=self.zmq_context())
        self.server.run_as_thread()
        self.backend.register_rpcs(self.server)
        self.wait_for_rpc_server(self.server)

    def test_datetime_is_converted_to_unix_timestamp(self):
        medium = MediumAccesses(
            id="GTEST1", type="lpr",
            accesses=[
                Access(
                    category="filler-visitor-button", user="test",
                    start=datetime(2021, 12, 25, 0, 0, 0), end=datetime(2021, 12, 25, 23, 59, 59)
                )
            ]
        )
        self.backend_class.database.append(medium)
        gateway = Gateway(gate="entry", direction="in")
        request = AccessRequest(gateway=gateway, name=self.core.config.NAME, medium_type="lpr", id="GTEST1")
        response = self.rpc("backend", "auth", request=request.dict())
        medium = response["response"]["medium"]
        self.assertTrue(isinstance(medium["accesses"][0]["start"], (int, float)))
        self.assertTrue(isinstance(medium["accesses"][0]["end"], (int, float)))



from typing import Dict

from openmodule.core import init_openmodule, shutdown_openmodule, OpenModuleCore
from openmodule_test.health import HealthTestMixin


class OpenModuleCoreTestMixin(HealthTestMixin):
    """
    Mixin which creates a core, zmq, and health mixin
    config-priority: config_kwargs -> zmq_config -> base_config, afterwards always use core.config
    """

    init_kwargs: Dict = {}
    config_kwargs: Dict = {}
    base_config = None

    core: OpenModuleCore

    def get_config_kwargs(self):
        return self.config_kwargs

    def get_init_kwargs(self):
        return self.init_kwargs

    def _fake_config(self):
        conf = self.zmq_config(**self.get_config_kwargs())
        if self.base_config:
            for key in vars(self.base_config):
                if not key.startswith("_") and key not in vars(conf):
                    setattr(conf, key, getattr(self.base_config, key))
        return conf

    def setUp(self):
        super().setUp()
        self.init_kwargs.setdefault("sentry", False)
        self.init_kwargs.setdefault("dsgvo", False)
        self.core = init_openmodule(
            config=self._fake_config(),
            context=self.zmq_context(),
            **self.get_init_kwargs()
        )
        self.zmq_client.subscribe(b"healthz")
        self.wait_for_health(self.core.config.NAME)

    def tearDown(self):
        try:
            super().tearDown()
        finally:
            shutdown_openmodule()



import random
import string
import threading
import time
from contextlib import suppress
from functools import partial

from pydantic.main import BaseModel

from openmodule_test.zeromq import ZMQTestMixin


class _EmptyModel(BaseModel):
    pass


class RPCServerTestMixin(ZMQTestMixin):
    def wait_for_rpc_server(self, server):
        pending_channels = {}
        lock = threading.Lock()

        def handler(_, __, channel):
            with lock:
                with suppress(KeyError):
                    del pending_channels[channel]

        """
        waits until a rpc server is responding on all channels
        """
        channels = set(x[0] for x in server.handlers)
        random_type = "_test" + "".join(random.choices(string.ascii_letters, k=10))
        for channel in channels:
            server.register_handler(channel, random_type, _EmptyModel, _EmptyModel, partial(handler, channel=channel))
            pending_channels[channel] = None

        for x in range(self.zmq_client.startup_check_iterations):
            with lock:
                channels = list(pending_channels.keys())
            if channels:
                for channel in channels:
                    self.rpc(channel, random_type, {}, receive_response=False)
                time.sleep(self.zmq_client.startup_check_delay)
            else:
                break

        assert not pending_channels, "error during startup and connect"
